@@ -5,6 +5,9 @@ use std::convert::TryFrom;
 use std::io;
 use std::ops::{Deref, DerefMut};
 
+#[cfg(feature = "tensor")]
+use afarray::Array;
+
 use async_trait::async_trait;
 use bytes::Bytes;
 use destream::{de, en};
@@ -40,7 +43,7 @@ pub trait BlockData: de::FromStream<Context = ()> + Clone + Send + Sync + 'stati
     }
 
     async fn load<S: AsyncReadExt + Send + Unpin>(source: S) -> TCResult<Self> {
-        destream_json::de::read_from((), source)
+        tbon::de::read_from((), source)
             .map_err(|e| TCError::internal(format!("unable to parse saved block: {}", e)))
             .await
     }
@@ -49,8 +52,8 @@ pub trait BlockData: de::FromStream<Context = ()> + Clone + Send + Sync + 'stati
     where
         Self: en::ToStream<'en>,
     {
-        let encoded = destream_json::en::encode(self)
-            .map_err(|e| TCError::internal(format!("unable to serialize Value: {}", e)))?;
+        let encoded = tbon::en::encode(self)
+            .map_err(|e| TCError::internal(format!("unable to serialize block: {}", e)))?;
 
         let mut reader = StreamReader::new(
             encoded
@@ -77,7 +80,7 @@ pub trait BlockData: de::FromStream<Context = ()> + Clone + Send + Sync + 'stati
     where
         Self: Clone + en::IntoStream<'en> + 'en,
     {
-        let encoded = destream_json::en::encode(self)
+        let encoded = tbon::en::encode(self)
             .map_err(|e| TCError::bad_request("serialization error", e))?;
 
         encoded
@@ -92,7 +95,7 @@ pub trait BlockData: de::FromStream<Context = ()> + Clone + Send + Sync + 'stati
     where
         Self: en::ToStream<'en>,
     {
-        let encoded = destream_json::en::encode(self)
+        let encoded = tbon::en::encode(self)
             .map_err(|e| TCError::bad_request("serialization error", e))?;
 
         encoded
@@ -115,6 +118,19 @@ impl BlockData for Value {
     }
 }
 
+#[cfg(feature = "tensor")]
+#[async_trait]
+impl BlockData for Array {
+    fn ext() -> &'static str {
+        "array"
+    }
+
+    #[inline]
+    fn max_size() -> u64 {
+        1_048_581 // 1 mebibyte + 5 bytes overhead
+    }
+}
+
 pub trait BlockRead<B: BlockData, F: File<B>>: Deref<Target = B> + Send {
     fn upgrade(self, file: &F)
         -> TCBoxTryFuture<<<F as File<B>>::Block as Block<B, F>>::WriteLock>;
@@ -206,12 +222,12 @@ pub trait File<B: BlockData>: Store + Sized + 'static {
 
 /// A transactional directory
 #[async_trait]
-pub trait Dir: Store + Sized + 'static {
+pub trait Dir: Store + Send + Sized + 'static {
     /// The type of a file entry in this `Dir`
     type File: Send;
 
     /// The `Class` of a file stored in this `Dir`
-    type FileClass;
+    type FileClass: Send;
 
     /// Return `true` if this directory has an entry at the given [`PathSegment`].
     async fn contains(&self, txn_id: &TxnId, name: &PathSegment) -> TCResult<bool>;
@@ -254,47 +270,47 @@ pub trait Dir: Store + Sized + 'static {
 
 /// Defines how to load a persistent data structure from the filesystem.
 #[async_trait]
-pub trait Persist<D: Dir, T: Transaction<D>>: Sized {
+pub trait Persist<D: Dir>: Sized {
     type Schema;
     type Store: Store;
+    type Txn: Transaction<D>;
 
     /// Return the schema of this persistent state.
     fn schema(&self) -> &Self::Schema;
 
     /// Load a saved state from persistent storage.
-    async fn load(txn: &T, schema: Self::Schema, store: Self::Store) -> TCResult<Self>;
-}
-
-/// Defines how to restore persistent state from backup.
-#[async_trait]
-pub trait Restore<D: Dir, T: Transaction<D>>: Sized {
-    async fn restore(&self, backup: &Self, txn_id: TxnId) -> TCResult<()>;
+    async fn load(txn: &Self::Txn, schema: Self::Schema, store: Self::Store) -> TCResult<Self>;
 }
 
 /// Defines how to copy a base state from another instance, possibly a view.
 #[async_trait]
-pub trait CopyFrom<D: Dir, T: Transaction<D>, Instance>: Persist<D, T> {
+pub trait CopyFrom<D: Dir, I>: Persist<D> {
     /// Copy a new instance of `Self` from an existing instance.
     async fn copy_from(
-        instance: Instance,
-        store: <Self as Persist<D, T>>::Store,
-        txn_id: TxnId,
-    ) -> TCResult<Self>
-    where
-        Instance: 'async_trait;
+        instance: I,
+        store: <Self as Persist<D>>::Store,
+        txn: <Self as Persist<D>>::Txn,
+    ) -> TCResult<Self>;
+}
+
+/// Defines how to restore persistent state from backup.
+#[async_trait]
+pub trait Restore<D: Dir>: Persist<D> {
+    async fn restore(&self, backup: &Self, txn_id: TxnId) -> TCResult<()>;
 }
 
 /// Defines a standard hash for a persistent state.
 #[async_trait]
-pub trait Hash<'en> {
+pub trait Hash<'en, D: Dir> {
     type Item: en::IntoStream<'en> + Send + 'en;
+    type Txn: Transaction<D>;
 
-    async fn hash_hex(&'en self, txn_id: TxnId) -> TCResult<String> {
-        self.hash(txn_id).map_ok(|hash| hex::encode(hash)).await
+    async fn hash_hex(&'en self, txn: &'en Self::Txn) -> TCResult<String> {
+        self.hash(txn).map_ok(|hash| hex::encode(hash)).await
     }
 
-    async fn hash(&'en self, txn_id: TxnId) -> TCResult<Bytes> {
-        let mut data = self.hashable(txn_id).await?;
+    async fn hash(&'en self, txn: &'en Self::Txn) -> TCResult<Bytes> {
+        let mut data = self.hashable(txn).await?;
 
         let mut hasher = Sha256::default();
         while let Some(item) = data.try_next().await? {
@@ -305,14 +321,14 @@ pub trait Hash<'en> {
         Ok(Bytes::from(digest.to_vec()))
     }
 
-    async fn hashable(&'en self, txn_id: TxnId) -> TCResult<TCTryStream<'en, Self::Item>>;
+    async fn hashable(&'en self, txn: &'en Self::Txn) -> TCResult<TCTryStream<'en, Self::Item>>;
 }
 
 async fn hash_chunks<'en, T: en::IntoStream<'en> + 'en>(
     hasher: &mut Sha256,
     data: T,
 ) -> TCResult<()> {
-    let mut data = destream_json::en::encode(data).map_err(TCError::internal)?;
+    let mut data = tbon::en::encode(data).map_err(TCError::internal)?;
     while let Some(chunk) = data.try_next().map_err(TCError::internal).await? {
         hasher.update(&chunk);
     }
